/*
 * Copyright (c) 1999, 2000 SuSE GmbH Nuernberg, Germany.
 * Author: Thorsten Kukuk <kukuk@suse.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, and the entire permission notice in its entirety,
 *    including the disclaimer of warranties.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.
 *
 * ALTERNATIVELY, this product may be distributed under the terms of
 * the GNU Public License, in which case the provisions of the GPL are
 * required INSTEAD OF the above restrictions.  (This clause is
 * necessary due to a potential bad interaction between the GPL and
 * the restrictions contained in a BSD-style copyright.)
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#define _GNU_SOURCE

#include <grp.h>
#include <pwd.h>
#include <time.h>
#include <crypt.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <shadow.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <rpc/rpc.h>
#include <rpc/key_prot.h>
#define PAM_SM_AUTH
#include <security/pam_modules.h>

#include "public.h"

/* This module actually performs UNIX/shadow authentication

   If shadow support is available, attempt to perform authentication
   using shadow passwords. If shadow is not available, or user does not
   have a shadow password, fallback onto a normal UNIX authentication.
   If an secret key exists and "set_secrpc" is set, send private key
   to keyserv.
*/

extern int key_setnet (struct key_netstarg *arg);

static void
_cleanup_message (pam_handle_t *pamh __attribute__ ((unused)),
		  void *fl, int err __attribute__ ((unused)))
{
  if (fl)
    free (fl);
}

static void
__delete_secret_key (uid_t uid)
{
  char empty_key[HEXKEYBYTES + 1];
  uid_t saved_uid = geteuid ();

  seteuid (uid);
  memset (&empty_key, 0, HEXKEYBYTES + 1);
  key_setsecret (empty_key);
  seteuid (saved_uid);
}

static int
__set_secret_key (pam_handle_t *pamh, uid_t uid, char *password)
{
  struct key_netstarg net;
  char domain[MAXNETNAMELEN + 1];
  char netname[MAXNETNAMELEN + 1];

  getdomainname (domain, MAXNETNAMELEN);

  /* nomalization of the domain name for secure RPC without the ending
     dot of the NIS+ domain. */
  if (domain[strlen (domain) - 1] == '.')
    domain[strlen (domain) - 1] = '\0';

  if (uid == 0)
    {
      char hostname[MAXHOSTNAMELEN + 1];

      gethostname (hostname, MAXHOSTNAMELEN);
      snprintf (netname, MAXNETNAMELEN, "unix.%s@%s", hostname, domain);
    }
  else
    snprintf (netname, MAXNETNAMELEN, "unix.%d@%s", uid, domain);

  /* encrypt the secret key with the users password */
  if (!getsecretkey (netname, (char *) &net.st_priv_key, password))
    {
      char *mtmp = malloc (strlen (netname) + 30);
      if (!mtmp)
	return PAM_IGNORE;
#if 0
      /* Don't print this, maybe no SecureRPC is used ? */
      sprintf (mtmp, "Can't find %s's secret key", netname);
      pam_set_data (pamh, "pam_unix_auth_keylogin_msg", mtmp,
		    _cleanup_message);
#endif
      return PAM_IGNORE;
    }

  /* check, if we have encrypt the secret key */
  if (net.st_priv_key[0] == 0)
    {
      char *mtmp = malloc (strlen (netname) + 40);
      if (!mtmp)
	return PAM_IGNORE;
      sprintf (mtmp, "Secure-RPC password incorrect for %s", netname);
      pam_set_data (pamh, "pam_unix_auth_keylogin_msg", mtmp,
		    _cleanup_message);
      return PAM_IGNORE;
    }

  /* give unencrypted secret key to keyserv */
  net.st_pub_key[0] = 0;
  net.st_netname = (char *) &netname;
  /* give unencrypted secret key to keyserv */
  if (key_setnet (&net) < 0)
    {
      char *mtmp = malloc (strlen (netname) + 65);
      if (!mtmp)
	return PAM_IGNORE;
      sprintf (mtmp,
	     "Could not set %s's secret key,\nmaybe the keyserver is down?",
	       netname);
      pam_set_data (pamh, "pam_unix_auth_keylogin_msg", mtmp,
		    _cleanup_message);
      return PAM_IGNORE;
    }

  return PAM_IGNORE;
}

int
pam_sm_authenticate (pam_handle_t * pamh, int flags, int argc,
		     const char **argv)
{
  int retval;
  struct spwd *sp = NULL;
  struct passwd *pw = NULL;
  const char *name = NULL;
  char *p = NULL;
  const char *salt;
  int dont_delete_seckey = 1;
  uid_t save_uid = geteuid ();
  /* Flags for options */
  int secrpc_flag = 0;
  int nullok = 0;
  int not_set_pass = 0;
  int use_first_pass = 0;

  /* Parse parameters for module */
  for ( ; argc-- > 0; argv++)
    {
      if (strcasecmp (*argv, "set_secrpc") == 0)
	secrpc_flag = 1;
      else if (strcasecmp (*argv, "nullok") == 0)
	nullok = 1;
      else if (strcasecmp (*argv, "not_set_pass") == 0)
	not_set_pass = 1;
      else if (strcasecmp (*argv, "use_first_pass") == 0)
	use_first_pass = 1;
      else
	__pam_log (LOG_ERR, "auth: Unknown option: %s", *argv);
    }


  /* get the user name */
  if ((retval = pam_get_user (pamh, &name, NULL)) != PAM_SUCCESS)
    return retval;

  if (name == NULL || !isalnum(*name))
    {
      if (name)
        __pam_log (LOG_ERR, "bad username [%s]", name);
      return PAM_USER_UNKNOWN;
    }

  /* Check at first for empty password, then prompt for user password */
  pw = getpwnam (name);
  /* Get shadow entry. We don't bail out if user does not exists.
     Ask for an password in this case and bail out then. */
  if (pw)
    sp = getspnam (pw->pw_name);

  if ((pw && (pw->pw_passwd == NULL || strlen (pw->pw_passwd) == 0)) ||
      (sp && pw && strcmp (pw->pw_passwd, "x") == 0 &&
       (sp->sp_pwdp == NULL || strlen (sp->sp_pwdp) == 0)))
    {
      if (flags & PAM_DISALLOW_NULL_AUTHTOK || !nullok)
	return PAM_AUTH_ERR;
      return PAM_SUCCESS;
    }

  retval = pam_get_item (pamh, PAM_AUTHTOK, (void *) &p);
  if (retval != PAM_SUCCESS)
    return retval;

  if (p == NULL)
    {
      if (use_first_pass)
	return PAM_AUTHTOK_RECOVER_ERR;

      retval = __get_authtok (pamh, not_set_pass);
      if (retval != PAM_SUCCESS)
	return retval;

      /* We have to call pam_get_item() again because value of p should
         changed to a password. */
      pam_get_item (pamh, PAM_AUTHTOK, (void *) &p);
    }

  if (pw == NULL)
    return PAM_USER_UNKNOWN;

  /* If authtok is empty again, check if this is allowed */
  if (p == NULL)
    {
      if (flags & PAM_DISALLOW_NULL_AUTHTOK || !nullok)
	return PAM_AUTH_ERR;
      return PAM_SUCCESS;
    }

  /* For NIS+, root cannot get password for lesser user.
     So we change our uid to this of the user, try to set
     his secret key and get the password a second time.
     This only works if we are allowed ot set the secret key. */
  if (secrpc_flag)
    {
      if (seteuid (pw->pw_uid) < 0)
	{
	  __pam_log (LOG_ERR, "auth: seteuid(%d) faild", pw->pw_uid);
	  return PAM_PERM_DENIED;
	}
      dont_delete_seckey = key_secretkey_is_set();
      __set_secret_key (pamh, pw->pw_uid, p);

      if (strcmp (pw->pw_passwd, "*NP*") == 0)
	pw = getpwnam (name);

      if (sp && strcmp (sp->sp_pwdp, "*NP*") == 0)
	sp = getspnam (pw->pw_name);

      if (seteuid (save_uid) < 0)
	{
	  __pam_log (LOG_ERR, "auth: seteuid(%d) faild", save_uid);
	  if (!dont_delete_seckey)
	    __delete_secret_key (pw->pw_uid);
	  return PAM_PERM_DENIED;
	}
      /* No we should have the correct password. */

      /* Check a second time, if the password is empty. It could
	 have changed since we now had the correct permissions. */
      if ((pw->pw_passwd == NULL || strlen (pw->pw_passwd) == 0) ||
	  (sp && strcmp (pw->pw_passwd, "x") == 0 &&
	   (sp->sp_pwdp == NULL || strlen (sp->sp_pwdp) == 0)) || p == NULL)
	{
	  if (flags & PAM_DISALLOW_NULL_AUTHTOK || !nullok)
	    {
	      if (!dont_delete_seckey)
		__delete_secret_key (pw->pw_uid);
	      return PAM_AUTH_ERR;
	    }
	  return PAM_SUCCESS;
	}
    }


  if (pw->pw_passwd[0] != '!')
    {
      if (sp)
	salt = sp->sp_pwdp;
      else
	salt = pw->pw_passwd;
    }
  else
    {
      if (!dont_delete_seckey && secrpc_flag)
	__delete_secret_key (pw->pw_uid);
      return PAM_PERM_DENIED;
    }


  /* This is for HP-UX password aging (why couldn't they use shadow ?) */
  if (strchr (salt, ',') != NULL)
    {
      char *cp = alloca (strlen (salt) + 1);
      strcpy (cp, salt);
      salt = cp;
      cp = strchr (salt, ',');
      *cp = '\0';
    }

  if (strcmp (crypt (p, salt), salt) != 0)
    {
      if (!dont_delete_seckey && secrpc_flag)
	__delete_secret_key (pw->pw_uid);
      return PAM_AUTH_ERR;
    }

  return PAM_SUCCESS;
}

int
pam_sm_setcred (pam_handle_t *pamh, int flags,
		int argc __attribute__ ((unused)),
		const char **argv __attribute__ ((unused)))
{
  char *msg = NULL;

  pam_get_data (pamh, "pam_unix_auth_keylogin_msg", (const void **) &msg);

  if (msg != NULL)
    __write_message (pamh, flags, PAM_TEXT_INFO, msg);

  return PAM_SUCCESS;
}
