/*
 * Copyright (c) 1999, 2000 SuSE GmbH Nuernberg, Germany.
 * Author: Thorsten Kukuk <kukuk@suse.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, and the entire permission notice in its entirety,
 *    including the disclaimer of warranties.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.
 *
 * ALTERNATIVELY, this product may be distributed under the terms of
 * the GNU Public License, in which case the provisions of the GPL are
 * required INSTEAD OF the above restrictions.  (This clause is
 * necessary due to a potential bad interaction between the GPL and
 * the restrictions contained in a BSD-style copyright.)
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#define _GNU_SOURCE

#define PAM_SM_PASSWORD

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pwd.h>
#include <crypt.h>
#include <shadow.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <rpc/types.h>
#include <nss.h>
#include <rpcsvc/yp_prot.h>
#include <rpcsvc/ypclnt.h>

#include <security/_pam_macros.h>
#include <security/pam_modules.h>

#include "yppasswd.h"
#include "md5.h"
#include "public.h"
#include "passwd_nss.h"

#define MAX_LOCK_RETRIES 3 /* How often should we try to lock password file */
#define MAX_PASSWD_TRIES 3
#define OLD_PASSWORD_PROMPT "Old Password: "
#define NEW_PASSWORD_PROMPT "New password: "
#define AGAIN_PASSWORD_PROMPT "Re-enter new password: "

#define SCALE (24L*3600L)

static int __do_setpass (pam_handle_t *pamh, int flag, user_t *usr, int, int);
static user_t *__do_getpwnam (char *user);
static int __get_passwd (pam_handle_t *pamh, const char *msgs, char **passwd);
static void __free_user_t (user_t *data);
static char *md5_crypt (const char *password);
extern char *bigcrypt(const char *key, const char *salt);

PAM_EXTERN int
pam_sm_chauthtok (pam_handle_t *pamh, int flags, int argc, const char **argv)
{
  user_t *data;
  char *user, *oldpass, *newpass;
  int retval, trys;
  int nullok = 0;
  int not_set_pass = 0;
  int use_first_pass = 0;
  int use_authtok = 0;
  int use_md5 = 0;
  int use_bigcrypt = 0;

  /* Parse parameters for module */
  for ( ; argc-- > 0; argv++)
    {
      if (strcasecmp (*argv, "nullok") == 0)
        nullok = 1;
      else if (strcasecmp (*argv, "not_set_pass") == 0)
        not_set_pass = 1;
      else if (strcasecmp (*argv, "use_first_pass") == 0)
        use_first_pass = 1;
      else if (strcasecmp (*argv, "use_authtok") == 0)
	use_authtok = 1;
      else if (strcasecmp (*argv, "md5") == 0)
	{
	  use_bigcrypt = 0;
	  use_md5 = 1;
	}
      else if (strcasecmp (*argv, "bigcrypt") == 0)
	{
	  use_md5 = 0;
	  use_bigcrypt = 1;
	}
      else
        __pam_log (LOG_ERR, "password: Unknown option: %s", *argv);
    }

  retval = pam_get_item (pamh, PAM_USER, (const void **) &user);
  if (retval != PAM_SUCCESS)
    return retval;

  if (user == NULL || strlen (user) == 0)
    {
      /* The app is supposed to get us the username! */
      return PAM_USER_UNKNOWN;
    }

  if (flags & PAM_PRELIM_CHECK)
    {
      /* This indicates that the modules are being probed as to their
         ready status for altering the user's authentication token. If
         the module requires access to another system over some network
         it should attempt to verify it can connect to this system on
         receiving this flag. If a module cannot establish it is ready to
         update the user's authentication token it should return
         PAM_TRY_AGAIN, this information will be passed back to the
         application. */

      retval = PAM_SUCCESS;
      data = __do_getpwnam (user);

      if (data == NULL || data->service == NONE)
	retval = PAM_USER_UNKNOWN;
      else if (data->is_expired)
	retval = PAM_ACCT_EXPIRED;

      if (data)
	__free_user_t (data);
      return retval;
    }

  /* Now we have all the initial information we need from the app to
     set things up (we assume that getting the username succeeded...) */
  data = __do_getpwnam (user);
  if (data == NULL || data->service == NONE)
    {
      if (data)
	__free_user_t (data);
      return PAM_USER_UNKNOWN;
    }

  if ((flags & PAM_CHANGE_EXPIRED_AUTHTOK) || getuid () != 0)
    {
      if (data->is_tooearly)
	{
	  if (data->use_hp_aging)
	    __write_message (pamh, flags, PAM_TEXT_INFO,
			  "Sorry: less then %d weeks since the last change.",
			     data->hp_week);
	  else
	    __write_message (pamh, flags, PAM_TEXT_INFO,
			     "Sorry: less then %d days since the last change.",
			     data->spw.sp_min);
	  __free_user_t (data);
	  return PAM_AUTHTOK_ERR;
	}
      else if (data->is_expired)
	{
	  __free_user_t (data);
	  return PAM_ACCT_EXPIRED; /* If their account has expired, we
				      can't auth them to change their
				      password */
	}
    }
  if ((data->use_shadow || data->use_hp_aging) && !data->is_expiring &&
      (flags & PAM_CHANGE_EXPIRED_AUTHTOK))
    {
      __free_user_t (data);
      return PAM_SUCCESS;
    }

  retval = pam_get_item (pamh, PAM_OLDAUTHTOK, (const void **) &oldpass);
  if (retval != PAM_SUCCESS)
    {
      __free_user_t (data);
      return retval;
    }

  if (oldpass == NULL)
    {
      if (getuid () || data->service != LOCAL)
	{
	  /* If this is being run by root and we change a local password,
	     we don't need to get the old password. */
	  if (use_first_pass)
	    {
	      __free_user_t (data);
	      if (getuid () == 0 && data->service != LOCAL)
		__write_message (pamh, flags, PAM_ERROR_MSG,
				 "You can only change local passwords.");
	      return PAM_AUTHTOK_RECOVER_ERR;
	    }

	  retval = __get_passwd (pamh, OLD_PASSWORD_PROMPT, &oldpass);
	  if (retval != PAM_SUCCESS)
	    {
	      __free_user_t (data);
	      return retval;
	    }
	  if (oldpass == NULL)
	    {
	      __free_user_t (data);
	      return PAM_AUTH_ERR;
	    }

	  /* Empty password means: entry is locked */
	  if ((data->oldpassword == NULL || strlen (data->oldpassword) == 0)
	      && !nullok)
	    {
	      __free_user_t (data);
	      return PAM_AUTH_ERR;
	    }

	  if (strncmp (data->oldpassword,
		       crypt (oldpass, data->oldpassword), 13))
	    {
	      __free_user_t (data);
	      return PAM_AUTH_ERR;
	    }

	  if (!not_set_pass)
	    pam_set_item (pamh, PAM_OLDAUTHTOK, (void *) oldpass);
	}
      else
	oldpass = "";
    }

  /* If we haven't been given a password yet, prompt for one... */
  trys = 0;
  pam_get_item (pamh, PAM_AUTHTOK, (const void **) &newpass);

  while ((newpass == NULL) && (trys++ < MAX_PASSWD_TRIES) && !use_authtok)
    {
      const char *cmiscptr = NULL;

      retval = __get_passwd (pamh, NEW_PASSWORD_PROMPT, &newpass);
      if (retval != PAM_SUCCESS)
	{
	  __free_user_t (data);
	  return retval;
	}

      if (newpass)
	{
	  if (oldpass)
	    if (!strcmp (oldpass, newpass))
	      {
		cmiscptr = "You must choose a new password.";
		newpass = NULL;
	      }
	}
      else
	{
	  /* We want to abort the password change */
	  __write_message (pamh, flags, PAM_ERROR_MSG,
			   "Password change aborted");
	  __free_user_t (data);
	  return PAM_AUTHTOK_ERR;
	}

      if (cmiscptr != NULL)
	{
	  __write_message (pamh, flags, PAM_ERROR_MSG, cmiscptr);
	  if (newpass != NULL)
	    {
	      memset (newpass, '\0', strlen (newpass));
	      free (newpass);
	      newpass = NULL;
	    }
	  if (trys >= MAX_PASSWD_TRIES)
	    {
	      __free_user_t (data);
	      return PAM_AUTHTOK_ERR;
	    }
	}
      else
	{
	  char *new2;
	  retval = __get_passwd (pamh, AGAIN_PASSWORD_PROMPT, &new2);
	  if (retval != PAM_SUCCESS)
	    {
	      __free_user_t (data);
	      return retval;
	    }

	  if (new2 == NULL)
	    {			/* Aborting password change... */
	      __write_message (pamh, flags, PAM_ERROR_MSG,
			       "Password change aborted");
	      __free_user_t (data);
	      return PAM_AUTHTOK_ERR;
	    }

	  if (strcmp (newpass, new2) == 0)
	    {
	      memset (new2, '\0', strlen (new2));
	      free (new2);
	      new2 = NULL;
	      break;
	    }
	  else
	    {
	      __write_message (pamh, flags, PAM_ERROR_MSG,
			       "They don't match; try again.");
	      if (new2 != NULL)
		{
		  memset (new2, '\0', strlen (new2));
		  free (new2);
		  new2 = NULL;
		}
	      if (newpass != NULL)
		{
		  memset (newpass, '\0', strlen (newpass));
		  free (newpass);
		  newpass = NULL;
		}
	    }
	}
    }
  /* end while (newpass == NULL) */

  if (newpass == NULL)
    {
      __free_user_t (data);
      return PAM_AUTHTOK_ERR;	/* They didn't seem to enter the right password
				   for three tries - error */
    }

  /* If we don't support passwords longer 8 characters, truncate them */
  if (!use_md5 && !use_bigcrypt && strlen (newpass) > 8)
    newpass[8] = '\0';
  data->newpassword = newpass;
  data->oldclearpwd = oldpass;
  retval = __do_setpass (pamh, flags, data, use_md5, use_bigcrypt);

  if (retval < 0)
    {
      __free_user_t (data);
      __write_message (pamh, flags, PAM_ERROR_MSG,
		       "Error: Password NOT changed");
      return PAM_AUTHTOK_ERR;
    }
  else
    {
      if (!not_set_pass)
	pam_set_item (pamh, PAM_AUTHTOK, (void *) newpass);

      __free_user_t (data);
      __write_message (pamh, flags, PAM_TEXT_INFO, "Password changed");
      return PAM_SUCCESS;
    }
}

static char *
getnismaster (pam_handle_t *pamh, int flags)
{

  char *master, *domainname;
  int port, err;

  yp_get_default_domain (&domainname);

  if ((err = yp_master (domainname, "passwd.byname", &master)) != 0)
    {
      __write_message (pamh, flags, PAM_ERROR_MSG,
		       "can't find the master ypserver: %s",
		       yperr_string (err));
      return NULL;
    }
  port = getrpcport (master, YPPASSWDPROG, YPPASSWDPROC_UPDATE, IPPROTO_UDP);
  if (port == 0)
    {
      __write_message (pamh, flags, PAM_ERROR_MSG,
		       "yppasswdd not running on NIS master %s", master);
      return NULL;
    }
  if (port >= IPPORT_RESERVED)
    {
      __write_message (pamh, flags, PAM_ERROR_MSG,
		       "yppasswd daemon running on illegal port.");
      return NULL;
    }

  return master;
}

static int
__do_setpass (pam_handle_t *pamh, int flags, user_t *data, int use_md5,
	      int use_bigcrypt)
{
#define bin_to_ascii(c) ((c)>=38?((c)-38+'a'):(c)>=12?((c)-12+'A'):(c)+'.')
  int retval = -1;
  char salt[2];
  time_t tm;

  time (&tm);
  salt[0] = bin_to_ascii (tm & 0x3f);
  salt[1] = bin_to_ascii ((tm >> 6) & 0x3f);

  if (data->service == LOCAL)
    {
      int retries = 0;
      while (lckpwdf () && retries < MAX_LOCK_RETRIES)
        {
          sleep (1);
          ++retries;
        }

      if (retries == MAX_LOCK_RETRIES)
        {
	  __write_message (pamh, flags, PAM_ERROR_MSG,
			   "Canot lock password file: already locked");
        }
      else if (data->use_shadow)
	{
	  struct stat passwd_stat;
	  struct spwd *sp; /* shadow struct obtained from fgetspent() */
	  FILE *oldpf = NULL, *newpf = NULL;
	  int gotit;

	  /* Open the shadow file for reading. We can't use getspent and
	     friends here, because they go through the YP maps, too. */
	  if ((oldpf = fopen ("/etc/shadow", "r")) == NULL)
	    {
	      __write_message (pamh, flags, PAM_ERROR_MSG,
			       "Can't open /etc/shadow: %m");
	      goto error_shadow;
	    }
	  if (fstat (fileno (oldpf), &passwd_stat) < 0)
	    {
	      __write_message (pamh, flags, PAM_ERROR_MSG,
			       "Can't stat /etc/shadow: %m");
	      fclose (oldpf);
	      goto error_shadow;
	    }

	  /* Open a temp shadow file */
	  if ((newpf = fopen ("/etc/shadow.tmp", "w+")) == NULL)
	    {
	      __write_message (pamh, flags, PAM_ERROR_MSG,
			       "Can't open /etc/shadow.tmp: %m");
	      fclose (oldpf);
	      oldpf=0;
	      goto error_shadow;
	    }
	  chmod ("/etc/shadow.tmp", passwd_stat.st_mode);
	  chown ("/etc/shadow.tmp", passwd_stat.st_uid, passwd_stat.st_gid);

	  gotit = 0;

	  /* Loop over all passwd entries */
	  while ((sp = fgetspent (oldpf)) != NULL)
	    {
	      /* check if this is the uid we want to change. A few
		 sanity checks added for consistency. */
	      if (!gotit && strcmp (data->pwd.pw_name, sp->sp_namp) == 0)
		{
		  time_t now;

		  time(&now);
		  /* set the new passwd */
		  if (use_md5)
		    sp->sp_pwdp = md5_crypt (data->newpassword);
		  else if (use_bigcrypt)
		    sp->sp_pwdp = bigcrypt (data->newpassword, salt);
		  else
		    sp->sp_pwdp = crypt (data->newpassword, salt);
		  sp->sp_lstchg = (long int)now / (24L*3600L);
		  gotit = 1;
		}

	      /* write the passwd entry to tmp file */
	      if (putspent (sp, newpf) < 0)
		{
		  __write_message (pamh, flags, PAM_ERROR_MSG,
				 "Error while writing new password file: %m");
		  goto error_shadow;
		}
	    }
	  retval = 0;
	  unlink ("/etc/shadow.old");
	  link ("/etc/shadow", "/etc/shadow.old");
	  rename ("/etc/shadow.tmp", "/etc/shadow");
	error_shadow:
	  unlink ("/etc/shadow.tmp");
	  if (oldpf) fclose (oldpf);
	  if (newpf) fclose (newpf);
	}
      else
	{
	  struct stat passwd_stat;
	  struct passwd *pw; /* passwd struct obtained from fgetpwent() */
	  FILE *oldpf = NULL, *newpf = NULL;
	  int gotit;

	  if ((oldpf = fopen ("/etc/passwd", "r")) == NULL)
	    {
	      __write_message (pamh, flags, PAM_ERROR_MSG,
			       "Can't open %s: %m", "/etc/passwd");
	      goto error_passwd;
	    }
          if (fstat (fileno (oldpf), &passwd_stat) < 0)
            {
              __write_message (pamh, flags, PAM_ERROR_MSG,
                               "Can't stat /etc/passwd: %m");
              fclose (oldpf);
              goto error_passwd;
            }

	  /* Open a temp passwd file */
	  if ((newpf = fopen ("/etc/passwd.tmp", "w+")) == NULL)
	    {
	      __write_message (pamh, flags, PAM_ERROR_MSG,
			       "Can't open %s: %m", "/etc/passwd.tmp");
	      fclose (oldpf);
	      oldpf=0;
	      goto error_passwd;
	    }
          chmod ("/etc/passwd.tmp", passwd_stat.st_mode);
          chown ("/etc/passwd.tmp", passwd_stat.st_uid, passwd_stat.st_gid);

	  gotit = 0;

	  /* Loop over all passwd entries */
	  while ((pw = fgetpwent (oldpf)) != NULL)
	    {
	      /* check if this is the uid we want to change. A few
		 sanity checks added for consistency. */
	      if (data->pwd.pw_uid == pw->pw_uid &&
		  data->pwd.pw_gid == pw->pw_gid &&
		  !strcmp (data->pwd.pw_name, pw->pw_name) && !gotit)
		{
		  if (use_md5)
		    pw->pw_passwd = md5_crypt (data->newpassword);
		  else if (use_bigcrypt)
		    pw->pw_passwd = bigcrypt (data->newpassword, salt);
		  else
		    pw->pw_passwd = crypt (data->newpassword, salt);
		  gotit = 1;
		}

	      /* write the passwd entry to tmp file */
	      if (putpwent (pw, newpf) < 0)
		{
		  __write_message (pamh, flags, PAM_ERROR_MSG,
				 "Error while writing new password file: %m");
		  goto error_passwd;
		}
	    }
	  retval = 0;
	  unlink ("/etc/passwd.old");
	  link ("/etc/passwd", "/etc/passwd.old");
	  rename ("/etc/passwd.tmp", "/etc/passwd");
	error_passwd:
	  unlink ("/etc/passwd.tmp");
	  if (oldpf) fclose (oldpf);
	  if (newpf) fclose (newpf);
	}

      ulckpwdf ();
    }
  else if (data->service == YP)
    {
      struct yppasswd yppwd;
      CLIENT *clnt;
      char *master = getnismaster(pamh, flags);
      struct timeval TIMEOUT = {25, 0}; /* total timeout */
      int error, status;

      if (master == NULL)
	return retval;

      /* Initialize password information */
      memset (&yppwd, '\0', sizeof (yppwd));
      if (use_md5)
	yppwd.newpw.pw_passwd = md5_crypt (data->newpassword);
      else if (use_bigcrypt)
	yppwd.newpw.pw_passwd = bigcrypt (data->newpassword, salt);
      else
	yppwd.newpw.pw_passwd = crypt (data->newpassword, salt);
      yppwd.newpw.pw_name = data->pwd.pw_name;
      yppwd.newpw.pw_uid = data->pwd.pw_uid;
      yppwd.newpw.pw_gid = data->pwd.pw_gid;
      yppwd.newpw.pw_gecos = data->pwd.pw_gecos;
      yppwd.newpw.pw_dir = data->pwd.pw_dir;
      yppwd.newpw.pw_shell = data->pwd.pw_shell;
      yppwd.oldpass = data->oldclearpwd;

      __write_message (pamh, flags, PAM_TEXT_INFO,
		       "Changing NIS password for %s on %s.",
		       data->pwd.pw_name, master);

      clnt = clnt_create (master, YPPASSWDPROG, YPPASSWDVERS, "udp");
      clnt->cl_auth = authunix_create_default ();
      memset (&status, '\0', sizeof (status));
      error = clnt_call (clnt, YPPASSWDPROC_UPDATE,
			 (xdrproc_t) xdr_yppasswd, (caddr_t) &yppwd,
			 (xdrproc_t) xdr_int, (caddr_t) &status, TIMEOUT);
      if (error || status)
	{
	  if (error)
	    clnt_perrno (error);
	  else
	    __write_message (pamh, flags, PAM_ERROR_MSG,
			     "Error while changing the NIS password.");
	}
      else
	retval = 0;
    }
  else if (data->service == NISPLUS)
    {
      retval = npd_upd_pwd (pamh, flags, data);
      if (retval != 0)
	{
	  __write_message (pamh, flags, PAM_ERROR_MSG,
			   "Error while changing the NIS+ password.");
	  retval = -1;
	}
    }

  return retval;
}

static void
__free_user_t (user_t *data)
{
  if (data->pwdbuffer)
    free (data->pwdbuffer);
  if (data->spwbuffer)
    free (data->spwbuffer);
  if (data->newpassword)
    free (data->newpassword);
  free (data);
}

/****************************************************************/

extern enum nss_status
_nss_files_getpwnam_r (const char *name, struct passwd *pwd,
		       char *buffer, size_t buflen, int *errnop);
extern enum nss_status
_nss_files_getspnam_r (const char *name, struct spwd *sp,
		       char *buffer, size_t buflen, int *errnop);

extern enum nss_status
_nss_nis_getpwnam_r (const char *name, struct passwd *pwd,
		     char *buffer, size_t buflen, int *errnop);
extern enum nss_status
_nss_nis_getspnam_r (const char *name, struct spwd *sp,
		     char *buffer, size_t buflen, int *errnop);

extern enum nss_status
_nss_nisplus_getpwnam_r (const char *name, struct passwd *pwd,
			 char *buffer, size_t buflen, int *errnop);
extern enum nss_status
_nss_nisplus_getspnam_r (const char *name, struct spwd *sp,
			 char *buffer, size_t buflen, int *errnop);

static user_t *
__do_getpwnam (char *user)
{
  enum nss_status status;
  user_t *data;
  struct nsw *nswp;
  int i;

  data = calloc (1, sizeof (user_t));
  if (data == NULL)
    return NULL;

  data->service = NONE;

  /* UNIX passwords area */

  nswp = _getnswbyname ("passwd");
  if (nswp == NULL)
    return data;

  for (i = 0; i < nswp->orderc; ++i)
    {
      const char *cmpptr = nswp->orders[i];

    again:

      if (nswp->orders[i][0] == '[')
	continue;
      if (strcmp ("files", cmpptr) == 0 || strcmp ("compat", cmpptr) == 0)
	{
	  /* Get password file entry... */
	  do {
	    data->pwdbuflen += 1024;
	    data->pwdbuffer = realloc (data->pwdbuffer, data->pwdbuflen);
	    status = _nss_files_getpwnam_r (user, &data->pwd, data->pwdbuffer,
					    data->pwdbuflen, &errno);
	  } while (status == NSS_STATUS_TRYAGAIN && errno == ERANGE);

	  if (status != NSS_STATUS_SUCCESS)
	    {
	      free (data->pwdbuffer);
	      data->pwdbuffer = NULL;
	      data->pwdbuflen = 0;
	      if (strcmp ("compat", cmpptr) == 0)
		{
		  struct nsw *nswp2 = _getnswbyname ("passwd_compat");

		  if (nswp2 == NULL)
		    cmpptr = "nis";
		  else
		    {
		      char *cp = alloca (strlen (nswp2->orders[0]) + 1);
		      strcpy (cp, nswp2->orders[0]);
		      cmpptr = cp;
		      nsw_free (nswp2);
		    }
		  goto again;
		}
	    }
	  else
	    {
	      data->service = LOCAL;
	      break;
	    }
	}
      else if (strcmp ("nis", cmpptr) == 0)
	{
	  /* Get password NIS entry... */
	  do {
	    data->pwdbuflen += 1024;
	    data->pwdbuffer = realloc (data->pwdbuffer, data->pwdbuflen);
	    status = _nss_nis_getpwnam_r (user, &data->pwd, data->pwdbuffer,
					  data->pwdbuflen, &errno);
	  } while (status == NSS_STATUS_TRYAGAIN && errno == ERANGE);

	  if (status != NSS_STATUS_SUCCESS)
	    {
	      free (data->pwdbuffer);
	      data->pwdbuffer = NULL;
	      data->pwdbuflen = 0;
	    }
	  else
	    {
	      data->service = YP;
	      break;
	    }
	}
      else if (strcmp ("nisplus", cmpptr) == 0)
	{
	  /* Get password NIS entry... */
	  do {
	    data->pwdbuflen += 1024;
	    data->pwdbuffer = realloc (data->pwdbuffer, data->pwdbuflen);
	    status = _nss_nisplus_getpwnam_r (user, &data->pwd,
					      data->pwdbuffer,
					      data->pwdbuflen, &errno);
	  } while (status == NSS_STATUS_TRYAGAIN && errno == ERANGE);

	  if (status != NSS_STATUS_SUCCESS)
	    {
	      free (data->pwdbuffer);
	      data->pwdbuffer = NULL;
	      data->pwdbuflen = 0;
	    }
	  else
	    {
	      data->service = NISPLUS;
	      break;
	    }
	}
    }

  nsw_free (nswp);

  if (data->service == NONE)
    return data;

  if (data->service == LOCAL)
    {
      do {
	data->spwbuflen += 1024;
	data->spwbuffer = realloc (data->spwbuffer, data->spwbuflen);
	status = _nss_files_getspnam_r (user, &data->spw, data->spwbuffer,
					data->spwbuflen, &errno);
      } while (status == NSS_STATUS_TRYAGAIN && errno == ERANGE);
    }
  else if (data->service == YP)
    {
      do {
	data->spwbuflen += 1024;
	data->spwbuffer = realloc (data->spwbuffer, data->spwbuflen);
	status = _nss_nis_getspnam_r (user, &data->spw, data->spwbuffer,
				      data->spwbuflen, &errno);
      } while (status == NSS_STATUS_TRYAGAIN && errno == ERANGE);
    }
  else if (data->service == NISPLUS)
    {
      do {
	data->spwbuflen += 1024;
	data->spwbuffer = realloc (data->spwbuffer, data->spwbuflen);
	status = _nss_nisplus_getspnam_r (user, &data->spw, data->spwbuffer,
					  data->spwbuflen, &errno);
      } while (status == NSS_STATUS_TRYAGAIN && errno == ERANGE);
    }
  else
    status = NSS_STATUS_NOTFOUND;

  if (status == NSS_STATUS_SUCCESS)
    data->use_shadow = TRUE;
  else if (strchr (data->pwd.pw_passwd, ',') != NULL)
    {
      char *age = strchr (data->pwd.pw_passwd, ',');
      long now = time (NULL) / (SCALE * 7);

      data->use_hp_aging = TRUE;
      *age++ = '\0';
      data->hp_max = c2n (age[0]);
      ++age;
      data->hp_min = c2n (age[0]);
      ++age;
      data->hp_week = str2week (age);

      if (data->hp_min > 0 && now <= (data->hp_week + data->hp_min))
	data->is_tooearly = TRUE;
      else if ((data->hp_max == 0 && data->hp_min == 0) ||
	       ((now > data->hp_week + data->hp_max) &&
		(data->hp_max >= data->hp_min)))
	data->is_expiring = TRUE;
    }

  if (data->use_shadow &&
      data->pwd.pw_passwd && data->pwd.pw_passwd[0] != '!')
    {
      long now = time (NULL) / SCALE;

      data->oldpassword = data->spw.sp_pwdp;

      if (data->spw.sp_lstchg > 0)
	{
	  if (data->spw.sp_min > 0 &&
	      now <= (data->spw.sp_lstchg + data->spw.sp_min))
	    data->is_tooearly = TRUE;
	  else if (data->spw.sp_inact >= 0 && data->spw.sp_max >= 0 &&
		   now >= (data->spw.sp_lstchg + data->spw.sp_max +
			   data->spw.sp_inact))
	    data->is_expired = TRUE;
	  else if (data->spw.sp_max >= 0 &&
		   now >= (data->spw.sp_lstchg + data->spw.sp_max))
	    data->is_expiring = TRUE;
	}
      else if (data->spw.sp_lstchg == 0)
	data->is_expiring = TRUE;
    }
  else
    data->oldpassword = data->pwd.pw_passwd;

  return data;
}

/* prompt user for a using conversation calls */
static int
__get_passwd (pam_handle_t *pamh, const char *msgs, char **passwd)
{
  struct pam_message msg[1], *pmsg[1];
  struct pam_response *resp;
  struct pam_conv *conv;
  int retval;

  /* set up conversation call */

  pmsg[0] = &msg[0];
  msg[0].msg_style = PAM_PROMPT_ECHO_OFF;
  msg[0].msg = msgs;
  resp = NULL;


  retval = pam_get_item (pamh, PAM_CONV, (const void **) &conv);
  if (retval == PAM_SUCCESS)
    {
      retval = conv->conv (1, (const struct pam_message **)pmsg,
                           &resp, conv->appdata_ptr);
      if (retval != PAM_SUCCESS)
        return retval;
    }
  else
    return retval;

  if (resp)
    {
      *passwd = strdup (resp->resp ? resp->resp : "");
      if (resp)
        _pam_drop_reply (resp, 1);
    }
  else
    return PAM_CONV_ERR;

  return PAM_SUCCESS;
}

/* i64c - convert an integer to a radix 64 character */
static int i64c (int i)
{
  if (i < 0)
    return ('.');
  else if (i > 63)
    return ('z');
  if (i == 0)
    return ('.');
  if (i == 1)
    return ('/');
  if (i >= 2 && i <= 11)
    return ('0' - 2 + i);
  if (i >= 12 && i <= 37)
    return ('A' - 12 + i);
  if (i >= 38 && i <= 63)
    return ('a' - 38 + i);
  return ('\0');
}

static char *md5_crypt (const char *password)
{
  static char buf[1024];
  struct timeval tv;
  MD5_CTX ctx;
  unsigned char result[16];
  char *cp = (char *) result;
  unsigned char tmp[16];
  int i;

  MD5Init (&ctx);
  gettimeofday( &tv, (struct timezone *) 0);
  MD5Update (&ctx, (void *) &tv, sizeof tv);
  i = getpid ();
  MD5Update (&ctx, (void *) &i, sizeof i);
  i = clock ();
  MD5Update (&ctx, (void *) &i, sizeof i);
  MD5Update (&ctx, result, sizeof result);
  MD5Final (tmp, &ctx);
  strcpy (cp, "$1$");      /* magic for the MD5 */
  cp += strlen (cp);
  for (i = 0; i < 8; ++i)
    *cp++ = i64c (tmp[i] & 077);
  *cp = '\0';

  /* no longer need cleartext */
  cp = crypt_md5 (password, (const char *) result);
  strncpy (buf, cp, sizeof (buf));        /* put e in malloc()ed memory */
  _pam_overwrite (cp);                    /* clean up */
  buf[sizeof (buf) - 1] = '\0';

  return buf;
}
